from collections import defaultdict


class key_dependent_dict(defaultdict):
    def __init__(self, fun_with_param):
        super().__init__(None)
        self.fun = fun_with_param

    def __missing__(self, key):
        ret = self.fun(key)
        self[key] = ret
        return ret



from unittest import TestCase

from disjoint_set.main import DisjointSet


class TestDisjointSet(TestCase):
    def setUp(self) -> None:
        self.dset = DisjointSet()

    def test_initializes_value_for_absent_key(self):
        self.assertEqual(
            self.dset.find(1),
            1
        )

    def test_holds_and_compares_different_data_types(self):
        values = [
            "a", 1, (1, 2, 3)
        ]

        for a, b in zip(values, values[1:]):
            self.assertEqual(
                self.dset.connected(a, b),
                False,
            )

    def test_unites_correctly(self):
        self.dset.union(1, 2)
        self.dset.union(3, 4)
        self.dset.union(1, 6)
        self.dset.union(8, 2)

        self.assertEqual(
            self.dset.connected(6, 8),
            True,
        )

        self.assertEqual(
            self.dset.connected(8, 6),
            True,
        )

    def test_finds_correctly(self):
        self.dset.union(1, 2)
        self.dset.union(2, 3)

        self.assertEqual(
            self.dset.find(3),
            3
        )

        self.assertEqual(
            self.dset.find(1),
            3
        )



from collections import defaultdict

from .utils import key_dependent_dict


class DisjointSet:
    def __init__(self):
        self._data = key_dependent_dict(lambda x: x)

    def find(self, x):
        """
        Returns the representative member of the set to which x belongs, may be x itself.
        :param x: element
        :return: representative
        """
        if x != self._data[x]:
            self._data[x] = self.find(self._data[x])
        return self._data[x]

    def union(self, x, y):
        """
        Attaches the roots of x and y trees together if they are not the same already.
        :param x: first element
        :param y: second element
        :return: None
        """
        parent_x, parent_y = self.find(x), self.find(y)
        if parent_x != parent_y:
            self._data[parent_x] = parent_y

    def connected(self, x, y):
        """
        Returns True if x and y belong to the same tree (i.e. they have the same root), False otherwise.
        :param x: first element
        :param y: second element
        :return: bool
        """
        return self.find(x) == self.find(y)



import setuptools

setuptools.setup(
    name='disjoint_set',
    version='0.0.2',
    packages=setuptools.find_packages(),
    description='Disjoint set data structure implementation for Python',
    url='https://github.com/mrapacz/disjoint_set',
    long_description=open('README.md').read(),
    maintainer="Maciej Rapacz",
    maintainer_email="python.disjointset@gmail.com",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
